; Defined by user elsewhere
; =========================

.import _main
.import _nmi_sound
.import _irq

; iNES header
; ===========

.segment "HEADER"

INES_MAPPER = 0
INES_MIRROR = 1 ; 0 = horizontal mirroring, 1 = vertical mirroring
INES_SRAM   = 0 ; 1 = battery backed SRAM at $6000-7FFF

.byte 'N', 'E', 'S', $1A ; ID
.byte $02 ; 16k PRG bank count
.byte $01 ; 4k CHR bank count
.byte INES_MIRROR | (INES_SRAM << 1) | ((INES_MAPPER & $f) << 4)
.byte (INES_MAPPER & %11110000)
.byte $0, $0, $0, $0, $0, $0, $0, $0 ; padding

; CHR-ROM
; =======

.segment "CHR"
.incbin "chr0.chr" ; background tiles
.incbin "chr1.chr" ; sprites

; Vectors
; =======

.segment "VECTORS"
.word vector_nmi
.word vector_reset
.word vector_irq

.export vector_reset ; exported for crt0.s

; Generic static variables to avoid stack in C
; ============================================

.export _i, _j, _k, _l, _a, _b;

.segment "ZEROPAGE"
_i: .res 1
_j: .res 1
_k: .res 2
_l: .res 2
_a: .res 1
_b: .res 1

; Gamepad handling
; ================

.export _gamepad0
.export _gamepad1

.export _gamepad_poll

.segment "ZEROPAGE"
_gamepad0: .res 1
_gamepad1: .res 1

.segment "ASM"

.proc _gamepad_poll
    ; latch the current controller state
    lda #1
    sta $4016
    lda #0
    sta $4016
    .repeat 2, I
        ; store high bit in gamepad to mark end of read
        lda #%10000000
        sta _gamepad0 + I
        ; read 8 bits from controller port
        :
            lda $4016 + I
            ror
            ror _gamepad0 + I
            bcc :-
    .endrepeat
    rts
.endproc

; PPU handling
; ============

; PPU data
.export _ppu_oam
.export _ppu_nam
.export _ppu_pal
.export _ppu_send
.export _ppu_frame
.export _ppu_ctrl
.export _ppu_mask
.export _ppu_scroll_x
.export _ppu_scroll_y

; PPU functions
.export _ppu_render_off
.export _ppu_nmi_off
.export _ppu_nmi_on
.export _ppu_nmi_wait
.export _ppu_sprite_0_hit
.export _ppu_latch_h
.export _ppu_latch_v
.export _ppu_write
.export _debug_grey

.segment "OAM"
_ppu_oam:      .res 256 ; 256 bytes for OAM sprite buffer

.segment "BSS"
;_ppu_nam:      .res 256 ; 256 bytes for nametable change buffer
_ppu_nam:      .res 64  ; reduced size to make room for more RAM use
_ppu_pal:      .res 32  ; 32 bytes for palette buffer

.segment "ZEROPAGE"
_ppu_send:     .res 1 ; boolean to signal ppu_oam/ppu_nam/ppu_pal are ready to send to PPU
_ppu_frame:    .res 1 ; incremented at every NMI
_ppu_ctrl:     .res 1 ; PPU_CTRL $2000 register, updated at NMI (bit 7 ignored)
_ppu_mask:     .res 1 ; PPU_MASK $2001 register, updated at NMI
_ppu_scroll_x: .res 1 ; PPU X scroll position
_ppu_scroll_y: .res 1 ; PPU Y scroll position

.segment "ASM"

vector_nmi:
    ; save registers
    pha
    txa
    pha
    tya
    pha
    ; disable NMI to prevent double-NMI if code runs long
    lda _ppu_ctrl
    and #%01111011 ; clear NMI bit and force horizontal fill
    sta _ppu_ctrl ; save horizontal fill bit for later
    sta $2000
    ; upload new PPU data if ready
    lda _ppu_send
    beq @no_send
        lda #0
        sta $2001 ; disable rendering
        ; upload OAM via DMA
        lda #0
        sta $2003 ; set offset to 0
        lda #>_ppu_oam
        sta $4014 ; issue DMA from _ppu_oam
        ; upload palettes
        bit $2002
        lda #$3F
        sta $2006
        lda #$00
        sta $2006
        ldx #(256 - 32)
        :
            lda _ppu_pal-(256-32), X
            sta $2007
            inx
            bne :-
        ; upload nametable changes
        ldx #0
        @nametable_job:
            lda _ppu_nam, X
            beq @nametable_done
            ; store byte count in Y
            and #%01111111
            tay
            ; set horizontal or vertical fill
            lda _ppu_nam, X
            and #%10000000
            lsr
            lsr
            lsr
            lsr
            lsr
            ora _ppu_ctrl
            sta $2000
            ; latch address
            inx
            lda _ppu_nam, X
            sta $2006
            inx
            lda _ppu_nam, X
            sta $2006
            inx
            ; write bytes
            :
                lda _ppu_nam, X
                sta $2007
                inx
                dey
                bne :-
            jmp @nametable_job
        @nametable_done:
        ; clear address
        lda #0
        sta $2006
        sta $2006
        ; changes finished, enable rendering
        lda _ppu_mask
        sta $2001
        ; mark data sent
        lda #0
        sta _ppu_send
    @no_send:
    ; set scroll
    bit $2002
    lda _ppu_scroll_x
    sta $2005
    lda _ppu_scroll_y
    sta $2005
    ; run sound routine
    jsr _nmi_sound
    ; increment frame counter
    inc _ppu_frame
    ; re-enable NMI
    lda _ppu_ctrl
    ora #%10000000 ; set NMI bit
    sta $2000
    ; restore registers
    pla
    tay
    pla
    tax
    pla
    rti

.proc _ppu_render_off
    lda #0
    sta $2001
    rts
.endproc

.proc _ppu_nmi_off
    lda _ppu_ctrl
    and #%01111111 ; clear NMI bit
    sta $2000
    rts
.endproc

.proc _ppu_nmi_on
    lda _ppu_ctrl
    ora #%10000000 ; set NMI bit
    sta $2000
    rts
.endproc

.proc _ppu_nmi_wait
    lda _ppu_frame
    :
        cmp _ppu_frame
        beq :-
    rts
.endproc

.proc _ppu_sprite_0_hit
    :
        bit $2002
        bmi :- ; wait for vblank to end if in vblank
    :
        bit $2002
        bvc :- ; wait for sprite 0 hit
    rts
.endproc

.proc _ppu_latch_h
    pha
    ; turn off NMI and select horizontal fill
    lda _ppu_ctrl
    and #%01111011
    sta $2000
    pla
    bit $2002
    stx $2006
    sta $2006
    rts
.endproc

.proc _ppu_latch_v
    pha
    ; turn off NMI and select vertical fill
    lda _ppu_ctrl
    and #%01111011
    ora #%00000100
    sta $2000
    pla
    bit $2002
    stx $2006
    sta $2006
    rts
.endproc

.proc _ppu_write
    sta $2007
    rts
.endproc

.proc ppu_init
    ; any ppu variables that need a nonzero setup
    lda #%00001000
    sta _ppu_ctrl
    lda #%00011110
    sta _ppu_mask
    ; OAM
    lda #0
    ldx #$ff
    :
        sta _ppu_oam, x
        inx
        bne :-
    rts
.endproc

.proc _debug_grey
    lda _ppu_mask
    ora #%00000001
    sta $2001
    rts
.endproc

; IRQ handling
; ============

.segment "ASM"

vector_irq:
    ; save registers
    pha
    tax
    pha
    tay
    pha
    ; call IRQ
    jsr _irq
    ; restore registers
    pla
    tay
    pla
    tax
    pla
    rti

; Reset and startup
; =================

.segment "ASM"

.import crt0_start

vector_reset:
    ; standard startup
    sei       ; set interrupt flag (unnecessary, unless reset is called from code)
    cld       ; disable decimal mode
    ldx #$40
    stx $4017 ; disable APU IRQ
    ldx $ff
    txs       ; set up stack
    ldx #$00
    stx $2000 ; disable NMI
    stx $2001 ; disable render
    stx $4010 ; disable DPCM IRQ
    stx $4015 ; disable audio
    bit $2002 ; clear vblank flag
    ; wait for vblank
    :
        bit $2002
        bpl :-
    ; clear memory
    lda #$00
    tax
    :
        sta $0000, X
        sta $0100, X
        sta $0200, X
        sta $0300, X
        sta $0400, X
        sta $0500, X
        sta $0600, X
        sta $0700, X
        inx
        bne :-
    ; wait for second vblank
    :
        bit $2002
        bpl :-
    ; PPU is now warmed up, NES is ready to go!
    ; setup PPU related variables
    jsr ppu_init
    ; setup CRT
    jsr crt0_start
    ; void main()
    jsr _main
    ; has returned from main(), reset
    jmp vector_reset

; end of file